/*******************************************************************************
 *
 * string_t.c 
 *
 * Provides a data structure and supporting functions for scanning
 * strings. 
 *
 * Cheetah Web Browser
 * Copyright (C) 2001 Garett Spencley 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 *
 *******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "string_t.h"
#include "error.h"

#define DEF_BUF_SIZE 32

/*
 * create_string() - creates a string structure 
 * based on the argument passed. 
 */

string_t *create_string(char *string)
{
	string_t *result;

	result = (string_t *)malloc(sizeof(string_t));
	if(!result)
		return NULL;

	result->buffer = strdup(string);
	if(!result->buffer) {
		free(result);
		return NULL;
	}

	result->len = strlen(string);
	result->pos = 0;

	return result;
}

/* 
 * free_string() - dispose of a string
 */

void free_string(string_t *string) 
{
	free(string->buffer);
	free(string);
}

/* 
 * get_char() - return the next char
 */

int get_char(string_t *string)
{
	if(string->pos == string->len)
		return 0;

	return string->buffer[string->pos++];
}	

/* 
 * unget_char() - Put back a character
 */

void unget_char(string_t *string)
{
	string->pos = (string->pos - 1) < 0 ? 0 : string->pos - 1;
}

/*
 * is_valid_ident() - returns 1 if c is a valid
 * character for an identifier, 0 otherwise.
 */

int is_valid_ident(int c)
{
	static const char valid[] =	{
		'/', '_', '.', '-', 0
	};

	int i;

	/* All alphanumeric characters are valid
	 * identifiers */

	if(isalnum(c)) return 1;

	/* Plus some of our own */

	for(i = 0; valid[i]; i++)
		if(c == valid[i])
			return 1;

	return 0;
}

/*
 * get_ident() - return an identifier
 */

char *get_ident(string_t *string)
{
	char *ident, *tmp;
	int counter;
	int bufsize = DEF_BUF_SIZE;
	int init_pos;

	/* Skip over all spaces */

	init_pos = string->pos;

	while(isspace(string->buffer[string->pos]) && string->pos < string->len) 
		++(string->pos);

	/* If there is no identifier in the string return NULL
       and set the position back to it's original */

	if(string->pos == string->len) {
		string->pos = init_pos;
		return NULL;
	}

	/* Allocate a default buffer */

	ident = (char *)malloc(sizeof(char) * bufsize);
	if(!ident) 
		return NULL;

	/* Fill in our buffer */
 
	counter = 0;

	while(is_valid_ident(string->buffer[string->pos]) 
		&& string->pos < string->len) {

		/* Resize our buffer if necessary */

		if(counter >= bufsize) {

			tmp = strdup(ident);
			if(!tmp) {
				error("out of memory.\n");
				free(ident);
				return NULL;
			}

			bufsize += DEF_BUF_SIZE;

			ident = (char *)realloc(ident, sizeof(char) * bufsize);
			if(!ident) {
				error("out of memory.\n");
				free(tmp);
				return NULL;
			}

			strcpy(ident, tmp);
			free(tmp);
		}

		ident[counter++] = string->buffer[string->pos++];
	}
	ident[counter] = 0;

	return ident;
}

/*
 * get_string() - returns the next string literal 
 * 
 * Note: The resulting string will not include the
 * quotation marks
 */

char *get_string(string_t *string)
{
	char *result, *tmp;
	int bufsize = DEF_BUF_SIZE;
	int counter;
	int init_pos;

	/* Skip until next '"' */

	init_pos = string->pos;

	while(string->buffer[string->pos++] != '"' && string->pos < string->len);

	if(string->pos == string->len) {
		string->pos = init_pos;
		return NULL;
	}

	/* Allocate a default buffer */

	result = (char *)malloc(sizeof(char) * bufsize);
	if(!result)
		return NULL;

	/* Fill in the buffer */

	counter = 0;

	while(string->buffer[string->pos] != '"' && string->pos < string->len) {
	
		/* Resize the buffer if necessary */

		if(counter >= bufsize) {

			tmp = strdup(result);
			if(!tmp) {
				error("out of memory.\n");
				free(result);
				return NULL;
			}

			bufsize += DEF_BUF_SIZE;

			result = (char *)realloc(result, sizeof(char) * bufsize);
			if(!result) {
				error("out of memroy.\n");
				free(tmp);
				return NULL;
			}

			strcpy(result, tmp);
			free(tmp);
		}

		result[counter++] = string->buffer[string->pos++];
	}			
	result[counter] = 0;

 	/* skip over last '"' */

	if(string->pos < string->len)
		string->pos++;

	return result;
}

/* 
 * get_rest() - return the rest of the string
 */

char *get_rest(string_t *string)
{
	char *result;
	int counter;

	result = (char *)malloc(sizeof(char) * string->len - string->pos + 1);
	if(!result)
		return NULL;
	
	counter = 0;

	for(; string->pos < string->len; string->pos++) 
		result[counter++] = string->buffer[string->pos];

	result[counter] = 0;

	return result;
}
